from django import forms


class ActionItemCrfFormMixin:

    """Declare with forms.ModelForm."""

    class Meta:
        help_text = {"action_identifier": "(read-only)"}
        widgets = {
            "action_identifier": forms.TextInput(
                attrs={"required": False, "readonly": "readonly"}
            ),
        }



from django import forms

from ..exceptions import ActionError


class ActionItemModelFormMixin:
    def clean(self):
        cleaned_data = super().clean()
        action_cls = self._meta.model.get_action_cls()
        if self.action_cls:
            try:
                action_cls(
                    subject_identifier=self.subject_identifier,
                    action_identifier=self.action_identifier,
                    related_action_item=self.related_action_item,
                    readonly=True,
                )
            except ActionError as e:
                raise forms.ValidationError(
                    f"{str(e.message)}. Please contact your data manager."
                )
        return cleaned_data

    @property
    def action_cls(self):
        try:
            action_cls = self._meta.model.get_action_cls()
        except AttributeError:
            action_cls = None
        return action_cls

    @property
    def action_identifier(self):
        action_identifier = self.cleaned_data.get("action_identifier")
        if not action_identifier:
            try:
                action_identifier = self.instance.action_identifier
            except AttributeError:
                action_identifier = None
        return action_identifier

    @property
    def related_action_item(self):
        related_action_item = self.cleaned_data.get("related_action_item")
        if not related_action_item:
            try:
                related_action_item = self.instance.related_action_item
            except AttributeError:
                related_action_item = None
        return related_action_item



from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase
from edc_constants.constants import CLOSED, NEW, NO, YES

from edc_action_item.get_action_type import get_action_type
from edc_action_item.helpers import ActionItemHelper
from edc_action_item.models import ActionItem, ActionType
from edc_action_item.site_action_items import site_action_items

from ..action_items import (
    FormFourAction,
    FormOneAction,
    FormThreeAction,
    FormTwoAction,
    FormZeroAction,
    SingletonAction,
    register_actions,
)
from ..models import FormFour, FormOne, FormThree, FormTwo, FormZero
from ..test_case_mixin import TestCaseMixin


class TestAction(TestCaseMixin, TestCase):
    def setUp(self):
        register_actions()
        self.subject_identifier = self.fake_enroll()
        self.assertIn(FormOneAction.name, site_action_items.registry)
        self.assertIn(FormTwoAction.name, site_action_items.registry)
        self.assertIn(FormThreeAction.name, site_action_items.registry)

    def test_str(self):
        action = FormZeroAction(subject_identifier=self.subject_identifier)
        self.assertTrue(str(action))

    def test_create_or_update_action_type(self):
        site_action_items.updated_action_types = False
        site_action_items.create_or_update_action_types()
        self.assertGreater(len(site_action_items.registry), 0)
        self.assertEqual(ActionType.objects.all().count(), len(site_action_items.registry))
        self.assertTrue(site_action_items.updated_action_types)

        site_action_items.create_or_update_action_types()
        self.assertGreater(len(site_action_items.registry), 0)
        self.assertEqual(ActionType.objects.all().count(), len(site_action_items.registry))

    def test_creates_own_action0(self):
        """Asserts a form creates its action item."""
        form_zero = FormZero.objects.create(subject_identifier=self.subject_identifier)
        try:
            ActionItem.objects.get(action_identifier=form_zero.action_identifier)
        except ObjectDoesNotExist:
            self.fail("Action item unexpectedly does not exist")
        for name in ["submit-form-zero"]:
            with self.subTest(name=name):
                try:
                    ActionItem.objects.get(
                        subject_identifier=self.subject_identifier,
                        action_type__name=name,
                    )
                except ObjectDoesNotExist:
                    self.fail("Action item unexpectedly does not exist.")
        self.assertEqual(
            ActionItem.objects.filter(subject_identifier=self.subject_identifier).count(),
            1,
        )

    def test_check_attrs_for_own_action0(self):
        """Test when model creates action."""
        obj = FormZero.objects.create(subject_identifier=self.subject_identifier)
        action_type = get_action_type(site_action_items.get(FormZero.action_name))
        action_item = ActionItem.objects.get(
            subject_identifier=self.subject_identifier,
            action_type__name="submit-form-zero",
        )
        self.assertEqual(action_item.subject_identifier, obj.subject_identifier)
        self.assertEqual(action_item.action_identifier, obj.action_identifier)
        self.assertEqual(action_item.action_identifier, obj.action_identifier)
        self.assertEqual(action_item.reference_model, obj._meta.label_lower)
        self.assertTrue(action_item.linked_to_reference)
        self.assertIsNone(action_item.related_action_item)
        self.assertIsNone(action_item.related_reference_model)
        self.assertIsNone(action_item.parent_action_item)
        # self.assertIsNone(action_item.parent_reference_model)
        self.assertIsNone(action_item.parent_action_item)
        self.assertEqual(action_item.action_type, action_type)

    def test_check_attrs_for_own_action1(self):
        """Test when action creates model."""
        obj = FormZero.objects.create(subject_identifier=self.subject_identifier)
        action = FormZeroAction(
            subject_identifier=self.subject_identifier,
            action_identifier=obj.action_identifier,
        )
        self.assertEqual(action.subject_identifier, obj.subject_identifier)
        self.assertEqual(action.action_identifier, obj.action_identifier)
        self.assertEqual(action.action_identifier, obj.action_identifier)
        self.assertEqual(action.reference_model, obj._meta.label_lower)
        self.assertTrue(action.linked_to_reference)
        self.assertIsNone(action.related_action_item)
        self.assertIsNone(action.related_reference_model)
        self.assertIsNone(action.parent_action_item)
        action_type = get_action_type(site_action_items.get(FormZero.action_name))
        self.assertEqual(get_action_type(action), action_type)

    def test_check_attrs_for_form_one_next_action(self):
        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)

        action_type_two = get_action_type(site_action_items.get(FormTwo.action_name))
        action_item_one = ActionItem.objects.get(
            action_identifier=form_one.action_identifier, status=CLOSED
        )
        action_item_two = ActionItem.objects.get(
            subject_identifier=self.subject_identifier,
            action_type__name="submit-form-two",
            status=NEW,
        )
        form_one = FormOne.objects.get(id=form_one.id)

        self.assertEqual(action_item_two.subject_identifier, form_one.subject_identifier)
        self.assertNotEqual(action_item_two.action_identifier, form_one.action_identifier)
        self.assertEqual(action_item_two.reference_model, FormTwo._meta.label_lower)
        self.assertEqual(action_item_two.related_action_item, form_one.action_item)
        self.assertEqual(action_item_two.related_reference_model, FormOne._meta.label_lower)
        self.assertEqual(action_item_two.parent_action_item, form_one.action_item)
        self.assertEqual(
            action_item_two.parent_action_item.reference_model,
            action_item_one.reference_model,
        )
        self.assertEqual(action_item_two.parent_action_item, action_item_one)
        self.assertEqual(get_action_type(action_item_two.action_cls), action_type_two)

    def test_does_not_duplicate_own_action_on_save(self):
        obj = FormZero.objects.create(subject_identifier=self.subject_identifier)
        obj.save()
        for name in ["submit-form-zero"]:
            with self.subTest(name=name):
                try:
                    ActionItem.objects.get(
                        subject_identifier=self.subject_identifier,
                        action_type__name=name,
                    )
                except ObjectDoesNotExist:
                    self.fail("Action item unexpectedly does not exist.")
        self.assertEqual(
            ActionItem.objects.filter(subject_identifier=self.subject_identifier).count(),
            1,
        )

    def test_creates_own_action1(self):
        obj = FormOne.objects.create(subject_identifier=self.subject_identifier)
        try:
            ActionItem.objects.get(action_identifier=obj.action_identifier)
        except ObjectDoesNotExist:
            self.fail("Action item unexpectedly does not exist")
        for name in ["submit-form-one", "submit-form-two", "submit-form-three"]:
            with self.subTest(name=name):
                try:
                    ActionItem.objects.get(
                        subject_identifier=self.subject_identifier,
                        action_type__name=name,
                    )
                except ObjectDoesNotExist:
                    self.fail("Action item unexpectedly does not exist.")
        self.assertEqual(
            ActionItem.objects.filter(subject_identifier=self.subject_identifier).count(),
            3,
        )

    def test_does_not_duplicate_own_actions_on_save(self):
        obj = FormOne.objects.create(subject_identifier=self.subject_identifier)
        self.assertEqual(
            ActionItem.objects.filter(subject_identifier=self.subject_identifier).count(),
            3,
        )
        obj.save()
        self.assertEqual(
            ActionItem.objects.filter(subject_identifier=self.subject_identifier).count(),
            3,
        )
        for name in ["submit-form-one", "submit-form-two", "submit-form-three"]:
            with self.subTest(name=name):
                try:
                    ActionItem.objects.get(
                        subject_identifier=self.subject_identifier,
                        action_type__name=name,
                    )
                except ObjectDoesNotExist:
                    self.fail("Action item unexpectedly does not exist.")

    def test_finds_existing_actions0(self):
        """Finds existing actions even when one is created in advance."""
        action_type = get_action_type(FormZeroAction)
        self.assertEqual(ActionItem.objects.all().count(), 0)
        ActionItem.objects.create(
            subject_identifier=self.subject_identifier, action_type=action_type
        )
        FormZero.objects.create(subject_identifier=self.subject_identifier)
        obj = FormZero.objects.get(subject_identifier=self.subject_identifier)
        self.assertTrue(ActionItem.objects.filter(action_identifier=obj.action_identifier))
        self.assertEqual(ActionItem.objects.all().count(), 1)
        obj.save()
        self.assertEqual(ActionItem.objects.all().count(), 1)

    def test_finds_existing_actions1(self):
        """Finds existing actions even when many are created in advance."""
        # create 5 action items for FormOne
        action_type = get_action_type(FormOneAction)
        self.assertEqual(ActionItem.objects.all().count(), 0)
        for _ in range(0, 5):
            ActionItem.objects.create(
                subject_identifier=self.subject_identifier, action_type=action_type
            )
        self.assertEqual(
            ActionItem.objects.filter(subject_identifier=self.subject_identifier).count(),
            5,
        )
        self.assertEqual(ActionItem.objects.filter(action_type=action_type).count(), 5)
        self.assertEqual(
            ActionItem.objects.filter(
                action_type=action_type,
                action_identifier__isnull=False,
                linked_to_reference=False,
            ).count(),
            5,
        )

        # create FormOne instances and expect them to link to
        # an exiting action item
        for i in range(0, 5):
            with self.subTest(index=i):
                obj = FormOne.objects.create(subject_identifier=self.subject_identifier)
                self.assertTrue(
                    ActionItem.objects.get(action_identifier=obj.action_identifier)
                )
                self.assertEqual(ActionItem.objects.filter(action_type=action_type).count(), 5)
                self.assertEqual(
                    ActionItem.objects.filter(
                        action_type=action_type, action_identifier=obj.action_identifier
                    ).count(),
                    1,
                )

    def test_finds_existing_actions2(self):
        action_type = get_action_type(FormOneAction)
        self.assertEqual(ActionItem.objects.all().count(), 0)
        for _ in range(0, 5):
            ActionItem.objects.create(
                subject_identifier=self.subject_identifier, action_type=action_type
            )
        self.assertEqual(ActionItem.objects.all().count(), 5)
        for _ in range(0, 5):
            FormOne.objects.create(subject_identifier=self.subject_identifier)
        self.assertEqual(ActionItem.objects.filter(action_type=action_type).count(), 5)
        self.assertEqual(
            ActionItem.objects.filter(
                action_type=action_type, action_identifier__isnull=True
            ).count(),
            0,
        )

    def test_creates_next_actions(self):
        f1_action_type = get_action_type(FormOneAction)
        f2_action_type = get_action_type(FormTwoAction)
        f3_action_type = get_action_type(FormThreeAction)
        FormOne.objects.create(subject_identifier=self.subject_identifier)
        self.assertEqual(ActionItem.objects.all().count(), 3)
        self.assertEqual(
            ActionItem.objects.filter(
                subject_identifier=self.subject_identifier, action_type=f1_action_type
            ).count(),
            1,
        )
        self.assertEqual(
            ActionItem.objects.filter(
                subject_identifier=self.subject_identifier, action_type=f2_action_type
            ).count(),
            1,
        )
        self.assertEqual(
            ActionItem.objects.filter(
                subject_identifier=self.subject_identifier, action_type=f3_action_type
            ).count(),
            1,
        )

    def test_creates_next_actions2(self):
        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        FormTwo.objects.create(subject_identifier=self.subject_identifier, form_one=form_one)
        FormThree.objects.create(subject_identifier=self.subject_identifier)

    def test_action_is_closed_if_model_creates_action(self):
        # form_one next_actions = [FormTwoAction, FormThreeAction]
        form_one_obj = FormOne.objects.create(subject_identifier=self.subject_identifier)
        self.assertEqual(ActionItem.objects.all().count(), 3)

        # next_actions = ['self']
        FormTwo.objects.create(
            subject_identifier=self.subject_identifier,
            parent_action_item=form_one_obj.action_item,
            form_one=form_one_obj,
        )
        self.assertEqual(ActionItem.objects.all().count(), 4)

        # next_actions = [FormZeroAction]
        # should find the existing and NEW FormThreeAction
        # instead of creating one.
        FormThree.objects.create(
            subject_identifier=self.subject_identifier,
            parent_action_item=form_one_obj.action_item,
        )
        self.assertEqual(ActionItem.objects.all().count(), 5)

        # 3 (1 for each model) are closed
        self.assertEqual(ActionItem.objects.filter(status=CLOSED).count(), 3)
        # next_actions are NEW
        self.assertEqual(ActionItem.objects.filter(status=NEW).count(), 2)

        f1_action_type = get_action_type(FormOneAction)
        f2_action_type = get_action_type(FormTwoAction)
        f3_action_type = get_action_type(FormThreeAction)

        obj = ActionItem.objects.get(
            subject_identifier=self.subject_identifier, action_type=f1_action_type
        )
        self.assertEqual(obj.status, CLOSED)

        obj = ActionItem.objects.get(
            subject_identifier=self.subject_identifier,
            action_type=f2_action_type,
            status=CLOSED,
        )
        self.assertEqual(obj.status, CLOSED)

        obj = ActionItem.objects.get(
            subject_identifier=self.subject_identifier,
            action_type=f3_action_type,
            status=CLOSED,
        )
        self.assertEqual(obj.status, CLOSED)

    def test_reference_url(self):
        action = FormOneAction(subject_identifier=self.subject_identifier)
        helper = ActionItemHelper(action_item=action.action_item)
        url = helper.get_url(model_cls=action.reference_model_cls())
        self.assertEqual(
            url,
            (
                f"/admin/edc_action_item/formone/add/?"
                f"action_identifier={action.action_identifier}"
            ),
        )

    def test_reference_url2(self):
        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        helper = ActionItemHelper(
            action_name=FormTwoAction.name, related_action_item=form_one.action_item
        )
        form_two_add_url = helper.reference_url
        self.assertEqual(
            form_two_add_url,
            f"/admin/edc_action_item/formtwo/add/?form_one={str(form_one.pk)}",
        )
        form_two = FormTwo.objects.create(
            subject_identifier=self.subject_identifier, form_one=form_one
        )
        helper = ActionItemHelper(action_item=form_two.action_item)
        form_two_change_url = helper.reference_url
        self.assertEqual(
            form_two_change_url,
            f"/admin/edc_action_item/formtwo/{str(form_two.pk)}/change/?"
            f"action_identifier={form_two.action_identifier}&form_one={str(form_one.pk)}",
        )

    def test_reference_url3(self):
        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        form_two_action = FormTwoAction(
            subject_identifier=self.subject_identifier,
            related_action_item=form_one.action_item,
        )
        helper = ActionItemHelper(action_item=form_two_action.action_item)
        self.assertEqual(
            helper.reference_url,
            (
                f"/admin/edc_action_item/formtwo/add/?"
                f"action_identifier={form_two_action.action_identifier}&"
                f"form_one={str(form_one.pk)}"
            ),
        )

    def test_create_singleton(self):
        action1 = SingletonAction(subject_identifier=self.subject_identifier)

        try:
            ActionItem.objects.get(subject_identifier=self.subject_identifier)
        except ObjectDoesNotExist:
            self.fail("ObjectDoesNotExist unexpectedly raised.")

        action2 = SingletonAction(subject_identifier=self.subject_identifier)

        self.assertEqual(action1.action_item, action2.action_item)

    def test_delete(self):
        """Assert action item is reset if reference object
        is deleted.
        """
        FormOneAction(subject_identifier=self.subject_identifier)
        action_item = ActionItem.objects.get(subject_identifier=self.subject_identifier)
        self.assertEqual(action_item.status, NEW)
        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        form_one = FormOne.objects.get(id=form_one.id)
        action_item = ActionItem.objects.get(
            subject_identifier=self.subject_identifier,
            action_identifier=form_one.action_identifier,
            action_type__name=FormOneAction.name,
        )
        self.assertEqual(action_item.status, CLOSED)
        form_one.delete()
        action_item = ActionItem.objects.get(
            subject_identifier=self.subject_identifier,
            action_identifier=form_one.action_identifier,
            action_type__name=FormOneAction.name,
        )
        self.assertEqual(action_item.status, NEW)

    def test_add_action_if_required(self):
        FormFourAction(subject_identifier=self.subject_identifier)
        form_four = FormFour.objects.create(
            subject_identifier=self.subject_identifier, happy=YES
        )

        self.assertRaises(
            ObjectDoesNotExist,
            ActionItem.objects.get,
            action_type=get_action_type(FormOneAction),
        )

        form_four.happy = NO
        form_four.save()
        try:
            ActionItem.objects.get(action_type=get_action_type(FormOneAction), status=NEW)
        except ObjectDoesNotExist:
            self.fail("action item unexpectedly does not exist")

        form_four.save()
        try:
            ActionItem.objects.get(action_type=get_action_type(FormOneAction), status=NEW)
        except ObjectDoesNotExist:
            self.fail("action item unexpectedly does not exist")

        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)

        form_four.save()
        try:
            ActionItem.objects.get(action_identifier=form_one.action_identifier, status=CLOSED)
        except ObjectDoesNotExist:
            self.fail("action item unexpectedly does not exist")



from edc_action_item.admin_site import edc_action_item_admin

from .models import CrfOne, CrfTwo, Followup, FormOne, FormTwo, Initial

edc_action_item_admin.register(CrfOne)
edc_action_item_admin.register(CrfTwo)
edc_action_item_admin.register(FormOne)
edc_action_item_admin.register(FormTwo)
edc_action_item_admin.register(Initial)
edc_action_item_admin.register(Followup)



# import sys
#
# from django.conf import settings
# from django.utils.safestring import mark_safe
# from django.core.mail.message import EmailMessage
# from edc_utils import get_utcnow
#
# NEW_REPORT = 'new_report'
# UPDATED_REPORT = 'updated_report'
#
# message_templates = {
#     NEW_REPORT: (
#         'Do not reply to this email\n\n'
#         '{test_message}'
#         'A report has been submitted for patient '
#         '{action_item.subject_identifier} '
#         'at site {action_item.site} which may require '
#         'your attention.\n\n'
#         'Title: {action_item.action_type.display_name}\n\n'
#         'You received this message because you are listed as a '
#         'member the Ambition Trial TMG\n\n'
#         '{test_message}'
#         'Thanks.'),
#     UPDATED_REPORT: (
#         'Do not reply to this email\n\n'
#         '{test_message}'
#         'A report has been updated for patient '
#         '{action_item.subject_identifier} '
#         'at site {action_item.site} which may require '
#         'your attention.\n\n'
#         'Title: {action_item.action_type.display_name}\n\n'
#         'Reason: {reason}\n\n'
#         'You received this message because you are listed as a '
#         'member the Ambition Trial TMG\n\n'
#         '{test_message}'
#         'Thanks.')
# }
#
#
# def send_email(action_item=None, reason=None, template_name=None,
#                force_send=None, using=None):
#     #     if template_name:
#     #         print('send_email', f'action_item={action_item}',
#     #               f'reason={reason}', f'template_name={template_name}')
#     #     else:
#     #         print('send_email', f'action_item={action_item}')
#     return None
#
#     if 'migrate' not in sys.argv:
#         template_name = template_name or 'new_report'
#         updated = '*UPDATE* ' if template_name == 'updated_report' else ''
#         email_recipients = action_item.action_cls.email_recipients
#         try:
#             email_enabled = settings.EMAIL_ENABLED
#         except AttributeError:
#             email_enabled = False
#         if email_enabled and email_recipients and (
#                 not action_item.emailed or force_send):
#             test_message = ''
#             test_subject = ''
#             try:
#                 live_system = settings.LIVE_SYSTEM
#             except AttributeError:
#                 live_system = False
#             if not live_system:
#                 test_message = 'THIS IS A TEST MESSAGE. NO ACTION IS REQUIRED\n\n'
#                 test_subject = 'TEST/UAT -- '
#             from_email = settings.EMAIL_CONTACTS.get('data_manager')
#             message = message_templates[template_name].format(
#                 test_message=test_message,
#                 action_item=action_item,
#                 reason=reason)
#             body = [mark_safe(message)]
#             email_message = EmailMessage(
#                 subject=(
#                     f'{test_subject}{updated}Ambition: '
#                     f'{action_item.action_type.display_name} '
#                     f'for {action_item.subject_identifier}'),
#                 body='\n\n'.join(body),
#                 from_email=from_email,
#                 to=email_recipients)
#             email_message.send()
#             action_item.emailed = True
#             action_item.emailed_datetime = get_utcnow()
#             action_item.save(using=using)



from .action import Action
from .site_action_items import site_action_items


class RegisterNotificationError(Exception):
    pass


def register(**kwargs):  # noqa
    """Registers a action_cls."""

    def _wrapper(action_cls):
        if not issubclass(action_cls, (Action,)):
            raise RegisterNotificationError(
                f"Wrapped class must a Action class. Got {action_cls}"
            )

        site_action_items.register(action_cls=action_cls)

        return action_cls

    return _wrapper



# Generated by Django 2.1.4 on 2019-01-14 00:50

import uuid

import _socket
import django.db.models.deletion
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import edc_utils
import simple_history.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("edc_action_item", "0014_auto_20181121_1738"),
    ]

    operations = [
        migrations.CreateModel(
            name="HistoricalReference",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                ("subject_identifier", models.CharField(max_length=50)),
                (
                    "parent_action_identifier",
                    models.CharField(
                        help_text="action identifier that links to parent reference model instance.",
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "related_action_identifier",
                    models.CharField(
                        help_text="action identifier that links to related reference model instance.",
                        max_length=30,
                        null=True,
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("action_identifier", models.CharField(db_index=True, max_length=25)),
                (
                    "report_datetime",
                    models.DateTimeField(default=edc_utils.date.get_utcnow),
                ),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                ("history_date", models.DateTimeField()),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "action_item",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_action_item.ActionItem",
                    ),
                ),
                (
                    "action_type",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_action_item.ActionType",
                        verbose_name="Action",
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "parent_action_item",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_action_item.ActionItem",
                    ),
                ),
                (
                    "related_action_item",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="edc_action_item.ActionItem",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical reference",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="action_type",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.ActionType",
                verbose_name="Action",
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="parent_action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.HistoricalActionItem",
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="related_action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.HistoricalActionItem",
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
    ]



from django import template
from edc_constants.constants import CANCELLED, CLOSED, HIGH_PRIORITY, NEW, OPEN
from edc_dashboard.utils import get_bootstrap_version

from ..choices import ACTION_STATUS
from ..helpers import ActionItemHelper
from ..site_action_items import site_action_items

register = template.Library()


@register.inclusion_tag(
    f"edc_action_item/bootstrap{get_bootstrap_version()}/add_action_item_popover.html"
)
def add_action_item_popover(subject_identifier, subject_dashboard_url) -> dict[str, str]:
    action_item_add_url = "edc_action_item_admin:edc_action_item_actionitem_add"
    show_link_to_add_actions = site_action_items.get_show_link_to_add_actions(
        subject_identifier
    )
    return dict(
        action_item_add_url=action_item_add_url,
        subject_identifier=subject_identifier,
        subject_dashboard_url=subject_dashboard_url,
        show_link_to_add_actions=show_link_to_add_actions,
    )


@register.inclusion_tag(
    f"edc_action_item/bootstrap{get_bootstrap_version()}/action_item_with_popover.html"
)
def action_item_with_popover(action_item_model_wrapper, tabindex):
    helper = ActionItemHelper(
        action_item=action_item_model_wrapper.object,
        href=action_item_model_wrapper.href,
    )
    context = helper.get_context()
    context.update(
        CANCELLED=[c[1] for c in ACTION_STATUS if c[0] == CANCELLED][0],
        CLOSED=[c[1] for c in ACTION_STATUS if c[0] == CLOSED][0],
        HIGH_PRIORITY=HIGH_PRIORITY,
        NEW=[c[1] for c in ACTION_STATUS if c[0] == NEW][0],
        OPEN=[c[1] for c in ACTION_STATUS if c[0] == OPEN][0],
        tabindex=tabindex,
    )
    return context



from django.test import TestCase
from django.urls.base import reverse
from django.views.generic.base import ContextMixin
from edc_model_wrapper import ModelWrapper

from edc_action_item.models import ActionItem, ActionType
from edc_action_item.templatetags.action_item_extras import add_action_item_popover
from edc_action_item.view_mixins import ActionItemViewMixin

from ..test_case_mixin import TestCaseMixin


class MyModelWrapper(ModelWrapper):
    next_url_name = "subject_dashboard_url"


class MyActionItemViewMixin(ActionItemViewMixin, ContextMixin):
    action_item_model_wrapper_cls = MyModelWrapper


class TestAction(TestCaseMixin, TestCase):
    def setUp(self):
        self.subject_identifier = self.fake_enroll()

    def test_view_context(self):
        view = MyActionItemViewMixin()
        view.kwargs = dict(subject_identifier=self.subject_identifier)
        context = view.get_context_data()
        self.assertEqual(context.get("open_action_items"), [])

        for action_type in ActionType.objects.all():
            ActionItem.objects.create(
                subject_identifier=self.subject_identifier, action_type=action_type
            )

        view = MyActionItemViewMixin()
        view.kwargs = dict(subject_identifier=self.subject_identifier)
        context = view.get_context_data()
        self.assertEqual(
            len(context.get("open_action_items")), ActionItem.objects.all().count()
        )

    def test_templatetag(self):
        context = add_action_item_popover(self.subject_identifier, "subject_dashboard_url")
        reverse(context.get("action_item_add_url"))



from django.apps import apps as django_apps
from django.test import TestCase
from edc_appointment.models import Appointment
from edc_facility import import_holidays
from edc_visit_tracking.constants import SCHEDULED

from edc_action_item import site_action_items
from edc_action_item.models import ActionItem

from ..action_items import CrfLongitudinalOneAction, CrfLongitudinalTwoAction
from ..models import CrfLongitudinalOne
from ..test_case_mixin import TestCaseMixin


class TestLongitudinal(TestCaseMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        site_action_items.registry = {}
        site_action_items.register(CrfLongitudinalOneAction)
        site_action_items.register(CrfLongitudinalTwoAction)
        self.subject_identifier = self.enroll()

    def test_(self):
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code="1000",
        )
        subject_visit = django_apps.get_model(
            "edc_visit_tracking.subjectvisit"
        ).objects.create(
            appointment=appointment,
            reason=SCHEDULED,
        )
        crf_one_a = CrfLongitudinalOne.objects.create(subject_visit=subject_visit)
        ActionItem.objects.get(action_identifier=crf_one_a.action_identifier)
        appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code="2000",
        )
        subject_visit = django_apps.get_model(
            "edc_visit_tracking.subjectvisit"
        ).objects.create(
            appointment=appointment,
            reason=SCHEDULED,
        )

        crf_one_b = CrfLongitudinalOne.objects.create(subject_visit=subject_visit)
        ActionItem.objects.get(action_identifier=crf_one_b.action_identifier)
        self.assertNotEqual(crf_one_a.action_identifier, crf_one_b.action_identifier)



from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist


def create_or_update_action_type(name=None, using=None, apps=None, **options):
    """Returns a model instance of ActionType.

    Gets or creates the ActionType on first pass.

    If model instance exists, updates.
    """

    action_type_model = "edc_action_item.actiontype"
    apps = apps or django_apps

    opts = {}
    action_type_model_cls = apps.get_model(action_type_model)
    if options:
        fields = [f.name for f in action_type_model_cls._meta.get_fields() if f.name != "name"]
        for attr, value in options.items():
            if attr in fields:
                opts.update({attr: value})
    try:
        action_type = action_type_model_cls.objects.get(name=name)
    except ObjectDoesNotExist:
        action_type = action_type_model_cls.objects.create(name=name, **opts)
    else:
        if opts:
            for k, v in opts.items():
                setattr(action_type, k, v)
            action_type.save(using=using)
            action_type = action_type_model_cls.objects.get(name=name)
    return action_type



from .action import Action
from .action_with_notification import ActionWithNotification
from .create_action_item import SingletonActionItemError, create_action_item
from .decorators import register
from .delete_action_item import ActionItemDeleteError, delete_action_item
from .fieldsets import action_fields, action_fieldset_tuple
from .modeladmin_mixins import ActionItemModelAdminMixin
from .site_action_items import site_action_items



from typing import Any

from django.db import models
from django.db.models.deletion import CASCADE, PROTECT
from edc_constants.choices import YES_NO
from edc_constants.constants import YES
from edc_crf.model_mixins import CrfWithActionModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.models import SiteModelMixin

from ..models import ActionModelMixin


class SubjectIdentifierModelManager(models.Manager):
    def get_by_natural_key(self, subject_identifier):
        return self.get(subject_identifier=subject_identifier)


class SubjectIdentifierModel(NonUniqueSubjectIdentifierFieldMixin, BaseUuidModel):
    objects = SubjectIdentifierModelManager()

    history = HistoricalRecords()

    def natural_key(self):
        return (self.subject_identifier,)  # noqa

    class Meta(BaseUuidModel.Meta, NonUniqueSubjectIdentifierFieldMixin.Meta):
        pass


class TestModelWithoutMixin(BaseUuidModel):
    subject_identifier = models.CharField(max_length=25)
    history = HistoricalRecords()


class TestModelWithActionDoesNotCreateAction(
    NonUniqueSubjectIdentifierFieldMixin,
    ActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    action_name = "test-nothing-prn-action"


class TestModelWithAction(
    NonUniqueSubjectIdentifierFieldMixin,
    ActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    action_name = "submit-form-zero"


class FormZero(
    NonUniqueSubjectIdentifierFieldMixin,
    ActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    action_name = "submit-form-zero"

    f1 = models.CharField(max_length=100, null=True)


class FormOne(
    NonUniqueSubjectIdentifierFieldMixin,
    ActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    action_name = "submit-form-one"

    f1 = models.CharField(max_length=100, null=True)


class FormTwo(
    NonUniqueSubjectIdentifierFieldMixin,
    ActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    form_one = models.ForeignKey(FormOne, on_delete=PROTECT)

    action_name = "submit-form-two"


class FormThree(
    NonUniqueSubjectIdentifierFieldMixin,
    ActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    action_name = "submit-form-three"


class FormFour(
    NonUniqueSubjectIdentifierFieldMixin,
    ActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    action_name = "submit-form-four"

    happy = models.CharField(max_length=10, choices=YES_NO, default=YES)


class Initial(
    NonUniqueSubjectIdentifierFieldMixin,
    ActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    action_name = "submit-initial"


class Followup(
    NonUniqueSubjectIdentifierFieldMixin,
    ActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    initial = models.ForeignKey(Initial, on_delete=CASCADE)

    action_name = "submit-followup"


class MyAction(
    NonUniqueSubjectIdentifierFieldMixin,
    ActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    action_name = "my-action"


class CrfOne(ActionModelMixin, SiteModelMixin, BaseUuidModel):
    subject_visit = models.OneToOneField(
        "edc_visit_tracking.subjectvisit",  # noqa
        on_delete=CASCADE,
        related_name="edc_action_item_test_visit_one",
    )

    action_name = "submit-crf-one"

    @property
    def subject_identifier(self: Any) -> str:
        return self.subject_visit.subject_identifier

    @property
    def related_visit(self):
        return getattr(self, self.related_visit_model_attr())

    @classmethod
    def related_visit_model_attr(cls):
        return "subject_visit"


class CrfTwo(ActionModelMixin, SiteModelMixin, BaseUuidModel):
    subject_visit = models.OneToOneField(
        "edc_visit_tracking.subjectvisit",  # noqa
        on_delete=CASCADE,
        related_name="edc_action_item_test_visit_two",
    )

    action_name = "submit-crf-two"

    @property
    def subject_identifier(self: Any) -> str:
        return self.subject_visit.subject_identifier

    @property
    def related_visit(self):
        return getattr(self, self.related_visit_model_attr())

    @classmethod
    def related_visit_model_attr(cls):
        return "subject_visit"


class CrfLongitudinalOne(
    CrfWithActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    action_name = "submit-crf-longitudinal-one"

    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)


class CrfLongitudinalTwo(
    CrfWithActionModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    action_name = "submit-crf-longitudinal-two"

    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)



# Generated by Django 2.2.6 on 2019-10-24 07:00

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_action_item", "0023_auto_20191007_0520")]

    operations = [
        migrations.AlterField(
            model_name="actionitem",
            name="site",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="sites.Site",
            ),
        )
    ]



ACTION_ITEM_COLUMNS = [
    "action_identifier",
    "action_item_id",
    "action_item_reason",
    "parent_action_identifier",
    "parent_action_item_id",
    "related_action_identifier",
    "related_action_item_id",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Type

from django.db import models
from django.db.models.deletion import PROTECT
from edc_model.models import HistoricalRecords

from ..exceptions import ActionClassNotDefined, ActionItemError
from ..managers import ActionIdentifierModelManager
from ..site_action_items import site_action_items
from .action_item import ActionItem

if TYPE_CHECKING:
    from ..action import Action


class ActionNoManagersModelMixin(models.Model):
    action_name: str = None

    action_item_model: str = "edc_action_item.actionitem"

    subject_dashboard_url: str = "subject_dashboard_url"

    action_identifier = models.CharField(max_length=50, unique=True, null=True, blank=True)

    action_item = models.ForeignKey(ActionItem, null=True, blank=True, on_delete=PROTECT)

    parent_action_item = models.ForeignKey(
        ActionItem, related_name="+", null=True, blank=True, on_delete=PROTECT
    )

    related_action_item = models.ForeignKey(
        ActionItem, related_name="+", null=True, blank=True, on_delete=PROTECT
    )

    # remove
    parent_action_identifier = models.CharField(
        max_length=30,
        null=True,
        blank=True,
        help_text="action identifier that links to parent reference model instance.",
    )

    # remove
    related_action_identifier = models.CharField(
        max_length=30,
        null=True,
        blank=True,
        help_text="action identifier that links to related reference model instance.",
    )

    action_item_reason = models.TextField(null=True, editable=False)

    def __str__(self) -> str:
        return f"{self.action_identifier[-9:]}"

    def save(self: Any, *args, **kwargs):
        # ensure action class is defined
        if not self.get_action_cls():
            raise ActionClassNotDefined(f"Action class name not defined. See {repr(self)}")

        # ensure subject_identifier
        if not self.subject_identifier:
            raise ActionItemError(
                f"Subject identifier may not be None. See {self.__class__}"
                f" action_identifier=`{self.action_identifier}`."
            )

        # ensure related_action_item is set if there is a
        # related reference model.
        if self.get_action_cls().related_reference_model and not self.related_action_item:
            self.related_action_item = getattr(
                self, self.get_action_cls().related_reference_fk_attr
            ).action_item

        if not self.id:
            # this is a new instance
            # associate a new or existing ActionItem
            # with this reference model instance
            action_cls = self.get_action_cls()
            action = action_cls(
                subject_identifier=self.subject_identifier,
                action_identifier=self.action_identifier,
                related_action_item=self.related_action_item,
            )
            self.action_item = action.action_item
            self.action_item.linked_to_reference = True
            self.action_identifier = self.action_item.action_identifier
        elif self.id and not self.action_item:
            self.action_item = ActionItem.objects.get(action_identifier=self.action_identifier)
        self.parent_action_item = self.action_item.parent_action_item

        # also see signals.py
        super().save(*args, **kwargs)  # type: ignore

    def natural_key(self: Any) -> tuple:
        return (self.action_identifier,)  # noqa

    # noinspection PyTypeHints
    natural_key.dependencies = ["edc_action_item.actionitem"]  # type:ignore

    @classmethod
    def get_action_cls(cls) -> Type[Action]:
        return site_action_items.get(cls.action_name)

    @property
    def action(self: Any):
        return self.get_action_cls()(
            subject_identifier=self.subject_identifier,
            action_item=self.action_item,
            readonly=True,
        )

    def get_action_item_reason(self):
        return self.action_item_reason or self.action_name

    @property
    def identifier(self):
        """Returns a shortened action_identifier"""
        return self.action_identifier[-9:]

    class Meta:
        abstract = True
        indexes = [
            models.Index(
                fields=[
                    "action_identifier",
                    "action_item",
                    "related_action_item",
                    "parent_action_item",
                ]
            )
        ]


class ActionModelMixin(ActionNoManagersModelMixin):
    objects = ActionIdentifierModelManager()

    history = HistoricalRecords(inherit=True)

    class Meta(ActionNoManagersModelMixin.Meta):
        abstract = True



# Generated by Django 4.2.7 on 2023-12-04 22:35

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_action_item", "0034_alter_actionitem_device_created_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="actionitem",
            options={
                "default_manager_name": "objects",
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
                "verbose_name": "Action Item",
                "verbose_name_plural": "Action Items",
            },
        ),
        migrations.AlterModelOptions(
            name="actiontype",
            options={
                "default_manager_name": "objects",
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
        ),
        migrations.AlterModelOptions(
            name="reference",
            options={
                "default_manager_name": "objects",
                "default_permissions": ("add", "change", "delete", "view", "export", "import"),
            },
        ),
        migrations.RemoveIndex(
            model_name="actionitem",
            name="edc_action__id_6cb4e0_idx",
        ),
        migrations.AddField(
            model_name="actionitem",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="actionitem",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="actiontype",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="actiontype",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="edcpermissions",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="edcpermissions",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalactionitem",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalactionitem",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalreference",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalreference",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="reference",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="reference",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddIndex(
            model_name="actionitem",
            index=models.Index(
                fields=["modified", "created"], name="edc_action__modifie_d8a9b7_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="actionitem",
            index=models.Index(
                fields=["user_modified", "user_created"], name="edc_action__user_mo_b01b34_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="actionitem",
            index=models.Index(
                fields=["subject_identifier"], name="edc_action__subject_0dae34_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="actionitem",
            index=models.Index(
                fields=[
                    "subject_identifier",
                    "action_identifier",
                    "action_type",
                    "site_id",
                    "status",
                    "report_datetime",
                ],
                name="edc_action__subject_5bacf2_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="actiontype",
            index=models.Index(
                fields=["modified", "created"], name="edc_action__modifie_7ecb65_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="actiontype",
            index=models.Index(
                fields=["user_modified", "user_created"], name="edc_action__user_mo_62771e_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="reference",
            index=models.Index(
                fields=["modified", "created"], name="edc_action__modifie_17f7a2_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="reference",
            index=models.Index(
                fields=["user_modified", "user_created"], name="edc_action__user_mo_e27ce1_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="reference",
            index=models.Index(
                fields=["subject_identifier"], name="edc_action__subject_b074e8_idx"
            ),
        ),
    ]



from typing import Any

from django.test import TestCase
from edc_constants.constants import CLOSED, NEW, OPEN
from edc_model_wrapper import ModelWrapper

from edc_action_item.models import ActionItem
from edc_action_item.site_action_items import site_action_items
from edc_action_item.templatetags.action_item_extras import action_item_with_popover

from ..action_items import (
    FormOneAction,
    FormThreeAction,
    FormTwoAction,
    register_actions,
)
from ..models import Followup, FormOne, FormTwo, Initial
from ..test_case_mixin import TestCaseMixin


class TestPopover(TestCaseMixin, TestCase):
    def setUp(self):
        register_actions()
        self.subject_identifier = self.fake_enroll()
        self.assertIn(FormOneAction.name, site_action_items.registry)
        self.assertIn(FormTwoAction.name, site_action_items.registry)
        self.assertIn(FormThreeAction.name, site_action_items.registry)

    def test_popover_templatetag(self):
        class ActionItemModelWrapper(ModelWrapper):
            model = "edc_action_item.actionitem"
            next_url_attrs = ["subject_identifier"]
            next_url_name = "subject_dashboard_url"

            @property
            def subject_identifier(self: Any):
                return self.object.subject_identifier

        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        obj = ActionItem.objects.get(action_identifier=form_one.action_identifier)
        wrapper = ActionItemModelWrapper(model_obj=obj)
        action_item_with_popover(wrapper, 0)
        context = action_item_with_popover(wrapper, 0)
        self.assertIsNone(context.get("parent_action_identifier"))
        self.assertIsNone(context.get("parent_action_item"))

        form_two = FormTwo.objects.create(
            subject_identifier=self.subject_identifier, form_one=form_one
        )
        obj = ActionItem.objects.get(action_identifier=form_two.action_identifier)
        wrapper = ActionItemModelWrapper(model_obj=obj)
        context = action_item_with_popover(wrapper, 0)
        self.assertEqual(context.get("parent_action_identifier"), form_one.action_identifier)
        self.assertEqual(context.get("parent_action_item"), form_one.action_item)

        context = action_item_with_popover(wrapper, 0)
        self.assertEqual(context.get("parent_action_identifier"), form_one.action_identifier)
        self.assertEqual(context.get("parent_action_item"), form_one.action_item)

    def test_popover_templatetag_action_url_if_reference_model_exists(self):
        """Asserts returns a change url if reference model
        exists.
        """

        class ActionItemModelWrapper(ModelWrapper):
            model = "edc_action_item.actionitem"
            next_url_attrs = ["subject_identifier"]
            next_url_name = "subject_dashboard_url"

            @property
            def subject_identifier(self: Any) -> str:
                return self.object.subject_identifier

        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        obj = ActionItem.objects.get(action_identifier=form_one.action_identifier)
        self.assertTrue(obj.status == CLOSED)
        obj.status = OPEN
        obj.save()
        wrapper = ActionItemModelWrapper(model_obj=obj)
        action_item_with_popover(wrapper, 0)
        context = action_item_with_popover(wrapper, 0)
        reference_url = context.get("reference_url")
        self.assertIn(f"{str(form_one.pk)}", reference_url)
        self.assertTrue(
            reference_url.startswith(
                f"/admin/edc_action_item/formone/{str(form_one.pk)}/change/"
            ),
            msg=reference_url,
        )

    def test_popover_templatetag2(self):
        class ActionItemModelWrapper(ModelWrapper):
            model = "edc_action_item.actionitem"
            next_url_attrs = ["subject_identifier"]
            next_url_name = "subject_dashboard_url"

            @property
            def subject_identifier(self: Any) -> str:
                return self.object.subject_identifier

        Initial.objects.create(subject_identifier=self.subject_identifier)
        Initial.objects.create(subject_identifier=self.subject_identifier)
        initial_obj = Initial.objects.create(subject_identifier=self.subject_identifier)
        initial_action_item_obj = ActionItem.objects.get(
            action_identifier=initial_obj.action_identifier
        )

        # assert initial action is correct
        self.assertEqual(
            initial_action_item_obj.subject_identifier, initial_obj.subject_identifier
        )
        self.assertEqual(
            initial_action_item_obj.action_identifier, initial_obj.action_identifier
        )
        self.assertIsNone(initial_action_item_obj.parent_action_item)
        self.assertEqual(
            initial_action_item_obj.action_identifier, initial_obj.action_identifier
        )

        # assert initial action is closed
        self.assertTrue(initial_action_item_obj.status == CLOSED)

        # assert followup action was created
        followup_action_obj = ActionItem.objects.get(
            parent_action_item_id=initial_action_item_obj.pk
        )

        # assert followup action is new
        self.assertTrue(followup_action_obj.status == NEW)

        # assert followup "add" url points to initial model obj fk
        wrapper = ActionItemModelWrapper(model_obj=followup_action_obj)
        action_item_with_popover(wrapper, 0)
        context = action_item_with_popover(wrapper, 0)
        reference_url = context.get("reference_url")
        self.assertIn("add", reference_url)
        self.assertIn(f"initial={str(initial_obj.pk)}", reference_url)

        # add a followup model instance
        followup_obj = Followup.objects.create(
            subject_identifier=initial_obj.subject_identifier,
            parent_action_item=initial_obj.action_item,
            initial=initial_obj,
        )

        # assert followup_obj has action identifier
        self.assertEqual(followup_obj.action_identifier, followup_action_obj.action_identifier)
        self.assertIsNotNone(followup_obj.action_identifier)

        # requery followup action
        followup_action_obj = ActionItem.objects.get(pk=followup_action_obj.pk)

        # assert followup 'change' url points to initial model obj
        followup_action_obj.status = OPEN
        followup_action_obj.save()
        wrapper = ActionItemModelWrapper(model_obj=followup_action_obj)
        action_item_with_popover(wrapper, 0)
        context = action_item_with_popover(wrapper, 0)
        reference_url = context.get("reference_url")
        # assert followup change url points to initial model obj
        self.assertIn("change", reference_url)
        self.assertIn(f"initial={str(initial_obj.pk)}", reference_url)

    def test_popover_templatetag3(self):
        class ActionItemModelWrapper(ModelWrapper):
            model = "edc_action_item.actionitem"
            next_url_attrs = ["subject_identifier"]
            next_url_name = "subject_dashboard_url"

            @property
            def subject_identifier(self: Any) -> str:
                return self.object.subject_identifier

        Initial.objects.create(subject_identifier=self.subject_identifier)

        initial_obj1 = Initial.objects.create(subject_identifier=self.subject_identifier)

        initial_obj2 = Initial.objects.create(subject_identifier=self.subject_identifier)

        followup_action_obj = ActionItem.objects.get(
            parent_action_item=initial_obj1.action_item
        )
        wrapper = ActionItemModelWrapper(model_obj=followup_action_obj)
        action_item_with_popover(wrapper, 0)
        context = action_item_with_popover(wrapper, 0)
        reference_url = context.get("reference_url")
        self.assertIn(f"initial={str(initial_obj1.pk)}", reference_url)

        followup_action_obj = ActionItem.objects.get(
            parent_action_item=initial_obj2.action_item
        )
        wrapper = ActionItemModelWrapper(model_obj=followup_action_obj)
        action_item_with_popover(wrapper, 0)
        context = action_item_with_popover(wrapper, 0)
        reference_url = context.get("reference_url")
        self.assertIn(f"initial={str(initial_obj2.pk)}", reference_url)

    def test_popover_templatetag4(self):
        class ActionItemModelWrapper(ModelWrapper):
            model = "edc_action_item.actionitem"
            next_url_attrs = ["subject_identifier"]
            next_url_name = "subject_dashboard_url"

            @property
            def subject_identifier(self: Any) -> str:
                return self.object.subject_identifier

        Initial.objects.create(subject_identifier=self.subject_identifier)

        initial_obj1 = Initial.objects.create(subject_identifier=self.subject_identifier)

        Initial.objects.create(subject_identifier=self.subject_identifier)

        followup_obj1 = Followup.objects.create(
            subject_identifier=initial_obj1.subject_identifier,
            parent_action_item=initial_obj1.action_item,
            initial=initial_obj1,
        )

        ActionItem.objects.get(
            subject_identifier=initial_obj1.subject_identifier,
            related_action_item=initial_obj1.action_item,
            parent_action_item=initial_obj1.action_item,
        )

        followup_action_obj2 = ActionItem.objects.get(
            subject_identifier=initial_obj1.subject_identifier,
            related_action_item=initial_obj1.action_item,
            parent_action_item=followup_obj1.action_item,
        )

        wrapper = ActionItemModelWrapper(model_obj=followup_action_obj2)
        action_item_with_popover(wrapper, 0)
        context = action_item_with_popover(wrapper, 0)
        reference_url = context.get("reference_url")
        self.assertIn(f"initial={str(initial_obj1.pk)}", reference_url)



from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase

from edc_action_item.get_action_type import get_action_type
from edc_action_item.models import ActionType
from edc_action_item.site_action_items import (
    AlreadyRegistered,
    SiteActionError,
    site_action_items,
)

from ..action_items import FormZeroAction
from ..test_case_mixin import TestCaseMixin


class TestSiteActionItems(TestCaseMixin, TestCase):
    def setUp(self):
        self.subject_identifier = self.fake_enroll()
        site_action_items.registry = {}
        get_action_type(FormZeroAction)
        self.action_type = ActionType.objects.get(name=FormZeroAction.name)

    def test_action_raises_if_not_registered(self):
        self.assertRaises(
            SiteActionError, FormZeroAction, subject_identifier=self.subject_identifier
        )

    def test_action_raises_if_already_registered(self):
        site_action_items.register(FormZeroAction)
        self.assertRaises(AlreadyRegistered, site_action_items.register, FormZeroAction)

    def test_action_instance_creates_action_type(self):
        ActionType.objects.all().delete()
        self.assertRaises(ObjectDoesNotExist, ActionType.objects.get, name=FormZeroAction.name)
        site_action_items.register(FormZeroAction)
        FormZeroAction(subject_identifier=self.subject_identifier)
        try:
            ActionType.objects.get(name=FormZeroAction.name)
        except ObjectDoesNotExist:
            self.fail("Object unexpectedly does not exist.")



from django.urls.conf import include, path
from edc_dashboard import url_names

from edc_action_item.admin_site import edc_action_item_admin

url_names.register("subject_dashboard_url", "subject_dashboard_url", "edc_action_item")


urlpatterns = [
    path("admin/", edc_action_item_admin.urls),
    path("", include("edc_action_item.urls")),
]



from django.core.exceptions import ObjectDoesNotExist
from django.db.models.deletion import ProtectedError
from django.test import TestCase
from edc_constants.constants import CANCELLED, CLOSED, NEW, OPEN

from edc_action_item.action import Action
from edc_action_item.create_or_update_action_type import create_or_update_action_type
from edc_action_item.exceptions import ActionItemStatusError, SubjectDoesNotExist
from edc_action_item.forms import ActionItemForm
from edc_action_item.get_action_type import get_action_type
from edc_action_item.models import ActionItem, ActionType
from edc_action_item.site_action_items import site_action_items

from ..action_items import FormOneAction, FormThreeAction, FormTwoAction, FormZeroAction
from ..models import FormOne, FormThree, FormTwo, TestModelWithAction
from ..test_case_mixin import TestCaseMixin


class TestActionItem(TestCaseMixin, TestCase):
    def setUp(self):
        self.subject_identifier = self.fake_enroll()
        site_action_items.registry = {}
        site_action_items.register(FormZeroAction)
        get_action_type(FormZeroAction)
        self.action_type = ActionType.objects.get(name=FormZeroAction.name)

    def test_creates(self):
        obj = ActionItem.objects.create(
            subject_identifier=self.subject_identifier,
            action_type=self.action_type,
            reference_model="edc_action_item.testmodel",
        )
        self.assertTrue(obj.action_identifier.startswith("AC"))
        self.assertEqual(obj.status, NEW)
        self.assertIsNotNone(obj.report_datetime)

    def test_create_requires_existing_subject(self):
        self.assertRaises(SubjectDoesNotExist, ActionItem.objects.create)

    def test_attrs(self):
        site_action_items.register(FormOneAction)
        site_action_items.register(FormTwoAction)
        site_action_items.register(FormThreeAction)
        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        form_one.refresh_from_db()
        form_two = FormTwo.objects.create(
            subject_identifier=self.subject_identifier, form_one=form_one
        )
        form_two.refresh_from_db()
        action_item_one = ActionItem.objects.get(action_identifier=form_one.action_identifier)
        action_item_two = ActionItem.objects.get(action_identifier=form_two.action_identifier)

        self.assertEqual(
            action_item_two.action_cls,
            site_action_items.get(action_item_two.action_type.name),
        )
        self.assertTrue(action_item_two.identifier)
        self.assertTrue(str(action_item_two))
        self.assertTrue(action_item_two.parent_action_item)
        self.assertIsNone(action_item_one.parent_action_item)

    def test_identifier_not_changed(self):
        site_action_items.registry = {}
        site_action_items.register(FormOneAction)
        get_action_type(FormOneAction)
        action_type = ActionType.objects.get(name=FormOneAction.name)
        obj = ActionItem.objects.create(
            subject_identifier=self.subject_identifier, action_type=action_type
        )
        action_identifier = obj.action_identifier
        obj.save()
        try:
            ActionItem.objects.get(action_identifier=action_identifier)
        except ObjectDoesNotExist:
            self.fail("ActionItem unexpectedly does not exist")

    def test_changes_action_item_status_from_new_to_open_on_edit(self):
        action_type = ActionType.objects.get(name=FormZeroAction.name)

        obj = ActionItem.objects.create(
            subject_identifier=self.subject_identifier, action_type=action_type
        )
        data = obj.__dict__
        data.update(action_type=obj.action_type.id)
        data["status"] = NEW
        form = ActionItemForm(data=obj.__dict__, instance=obj)
        form.is_valid()
        self.assertNotIn("status", form.errors)
        self.assertEqual(form.cleaned_data.get("status"), OPEN)

    def test_action_type_update_from_action_classes(self):
        class MyAction(Action):
            name = "my-action"
            display_name = "my action"
            reference_model = "edc_action_item.reference"

        class MyActionWithNextAction(Action):
            name = "my-action-with-next-as-self"
            display_name = "my action with next as self"
            next_actions = [MyAction]
            reference_model = "edc_action_item.reference"

        class MyActionWithNextActionAsSelf(Action):
            name = "my-action-with-next"
            display_name = "my action with next"
            next_actions = ["self"]
            reference_model = "edc_action_item.reference"

        site_action_items.register(MyAction)
        site_action_items.register(MyActionWithNextAction)
        site_action_items.register(MyActionWithNextActionAsSelf)
        my_action = MyAction(subject_identifier=self.subject_identifier)

        try:
            action_item = ActionItem.objects.get(action_identifier=my_action.action_identifier)
        except ObjectDoesNotExist:
            self.fail("ActionItem unexpectedly does not exist")

        self.assertEqual(my_action.action_item, action_item)
        self.assertEqual(my_action.action_identifier, action_item.action_identifier)
        self.assertEqual(get_action_type(my_action), action_item.action_type)
        self.assertEqual(
            get_action_type(my_action).reference_model, action_item.reference_model
        )
        self.assertIsNone(action_item.parent_action_item_id)

        class MyActionWithIncorrectModel(Action):
            name = "my-action2"
            display_name = "my action 2"
            reference_model = "edc_action_item.TestModelWithAction"

        site_action_items.register(MyActionWithIncorrectModel)

        TestModelWithAction.objects.create(subject_identifier=self.subject_identifier)
        self.assertRaises(
            ObjectDoesNotExist,
            TestModelWithAction.objects.create,
            subject_identifier=self.subject_identifier,
            action_identifier="blahblah",
        )

    def test_action_type_updates(self):
        class MyAction(Action):
            name = "my-action3"
            display_name = "original display_name"
            reference_model = "edc_action_item.FormOne"

        site_action_items.register(MyAction)
        MyAction(subject_identifier=self.subject_identifier)
        action_type = ActionType.objects.get(name="my-action3")
        self.assertEqual(action_type.display_name, "original display_name")

        site_action_items.registry = {}
        MyAction.display_name = "changed display_name"
        site_action_items.register(MyAction)
        create_or_update_action_type(**MyAction.as_dict())
        MyAction(subject_identifier=self.subject_identifier)
        action_type = ActionType.objects.get(name="my-action3")
        self.assertEqual(action_type.display_name, "changed display_name")

    def test_delete_child_and_parent_recreates(self):
        site_action_items.register(FormOneAction)
        site_action_items.register(FormTwoAction)
        site_action_items.register(FormThreeAction)
        form_one_action = FormOneAction(subject_identifier=self.subject_identifier)
        form_one = FormOne.objects.create(
            subject_identifier=self.subject_identifier,
            action_identifier=form_one_action.action_identifier,
        )
        action_item = ActionItem.objects.get(action_type__name=FormTwoAction.name)
        action_item.delete()
        action_item = ActionItem.objects.get(action_type__name=FormThreeAction.name)
        action_item.delete()
        # assert deleted actions are recreated
        self.assertEqual(ActionItem.objects.all().count(), 3)
        # assert two are parents of form one
        self.assertEqual(
            ActionItem.objects.filter(
                parent_action_item__action_identifier=form_one.action_identifier
            ).count(),
            2,
        )

    def test_delete2(self):
        site_action_items.register(FormOneAction)
        site_action_items.register(FormTwoAction)
        site_action_items.register(FormThreeAction)
        form_one_action = FormOneAction(subject_identifier=self.subject_identifier)
        form_one = FormOne.objects.create(
            subject_identifier=self.subject_identifier,
            action_identifier=form_one_action.action_identifier,
        )

        # delete form one
        form_one.delete()
        # assert resets action item
        action_item = ActionItem.objects.get(action_type__name=FormOneAction.name, status=NEW)

        # assert cleans up child action items
        self.assertRaises(
            ObjectDoesNotExist,
            ActionItem.objects.get,
            action_type__name=FormTwoAction.name,
        )
        self.assertRaises(
            ObjectDoesNotExist,
            ActionItem.objects.get,
            action_type__name=FormThreeAction.name,
        )

        # assert can delete form one action
        action_item.delete()

    def test_delete3(self):
        """Assert cannot delete action item if reference
        object exists.
        """
        site_action_items.register(FormOneAction)
        site_action_items.register(FormTwoAction)
        site_action_items.register(FormThreeAction)
        form_one_action = FormOneAction(subject_identifier=self.subject_identifier)
        FormOne.objects.create(
            subject_identifier=self.subject_identifier,
            action_identifier=form_one_action.action_identifier,
        )

        action_item = ActionItem.objects.get(action_type__name=FormOneAction.name)

        self.assertRaises(ProtectedError, action_item.delete)

    def test_new(self):
        site_action_items.register(FormOneAction)
        site_action_items.register(FormTwoAction)
        site_action_items.register(FormThreeAction)

        form_one_action = FormOneAction(subject_identifier=self.subject_identifier)

        self.assertEqual(form_one_action.action_item.status, NEW)

    def test_close_on_create(self):
        site_action_items.register(FormOneAction)
        site_action_items.register(FormTwoAction)
        site_action_items.register(FormThreeAction)

        form_one_action = FormOneAction(subject_identifier=self.subject_identifier)

        self.assertEqual(form_one_action.action_item.status, NEW)

        form_one = FormOne.objects.create(
            subject_identifier=self.subject_identifier,
            action_identifier=form_one_action.action_identifier,
        )

        self.assertEqual(form_one.action_item.status, CLOSED)

    def test_all_closed_on_create(self):
        site_action_items.register(FormOneAction)
        site_action_items.register(FormTwoAction)
        site_action_items.register(FormThreeAction)

        form_one_action = FormOneAction(subject_identifier=self.subject_identifier)

        form_one = FormOne.objects.create(
            subject_identifier=self.subject_identifier,
            action_identifier=form_one_action.action_identifier,
        )

        form_two = FormTwo.objects.create(
            subject_identifier=self.subject_identifier, form_one=form_one
        )

        form_three = FormThree.objects.create(subject_identifier=self.subject_identifier)

        self.assertEqual(form_one.action_item.status, CLOSED)
        self.assertEqual(form_two.action_item.status, CLOSED)
        self.assertEqual(form_three.action_item.status, CLOSED)

    def test_detects_change(self):
        site_action_items.register(FormOneAction)
        site_action_items.register(FormTwoAction)
        site_action_items.register(FormThreeAction)

        form_one_action = FormOneAction(subject_identifier=self.subject_identifier)
        form_one = FormOne.objects.create(
            subject_identifier=self.subject_identifier,
            action_identifier=form_one_action.action_identifier,
        )

        form_one.f1 = "blah"
        form_one.save()

        form_one.refresh_from_db()

        self.assertIsNotNone(form_one.action.reference_obj_has_changed)

    def test_reopens_children_on_change(self):
        site_action_items.register(FormOneAction)
        site_action_items.register(FormTwoAction)
        site_action_items.register(FormThreeAction)

        form_one_action = FormOneAction(subject_identifier=self.subject_identifier)
        form_one = FormOne.objects.create(
            subject_identifier=self.subject_identifier,
            action_identifier=form_one_action.action_identifier,
        )
        form_two = FormTwo.objects.create(
            subject_identifier=self.subject_identifier, form_one=form_one
        )
        form_three = FormThree.objects.create(subject_identifier=self.subject_identifier)

        form_one.f1 = "blah"
        form_one.save()

        form_one.refresh_from_db()
        form_two.refresh_from_db()
        form_three.refresh_from_db()

        self.assertEqual(form_two.action_item.status, OPEN)
        self.assertEqual(form_three.action_item.status, OPEN)
        self.assertEqual(form_one.action_item.status, CLOSED)

    def test_cannot_cancel_if_reference_obj_exists(self):
        site_action_items.register(FormOneAction)
        site_action_items.register(FormTwoAction)
        site_action_items.register(FormThreeAction)

        form_one_action = FormOneAction(subject_identifier=self.subject_identifier)
        form_one = FormOne.objects.create(
            subject_identifier=self.subject_identifier,
            action_identifier=form_one_action.action_identifier,
        )
        form_one.f1 = "blah"
        form_one.save()

        form_one.refresh_from_db()
        self.assertEqual(form_one.action_item.status, CLOSED)

        form_one.action_item.status = CANCELLED
        self.assertRaises(ActionItemStatusError, form_one.action_item.save)



import re
import sys

from django.core.exceptions import (
    FieldError,
    MultipleObjectsReturned,
    ObjectDoesNotExist,
)
from django.db import connection
from django.db.models.signals import post_delete, post_save, pre_save
from edc_constants.constants import CLOSED, NEW, OPEN

from .site_action_items import site_action_items


def fix_null_historical_action_identifier(app_label, models):
    """Fix null action_identifiers from previous versions."""
    with connection.cursor() as cursor:
        for model in models:
            tbl = f"{app_label}_historical{model}"
            if not re.match("([a-z_]+)_historical([a-z]+)", tbl):
                raise ValueError(f"Invalid table name when building sql statement. Got {tbl}")
            tbl = re.match("([a-z_]+)_historical([a-z]+)", tbl).group()
            sql = (
                f"update {tbl} set action_identifier=id "  # nosec B608
                "where action_identifier is null"  # nosec B608
            )
            cursor.execute(sql)


def fix_null_action_item_fk(apps, app_label, models):
    """Re-save instances to update action_item FKs."""
    action_item_cls = apps.get_model("edc_action_item", "ActionItem")
    post_save.disconnect(dispatch_uid="serialize_on_save")
    pre_save.disconnect(dispatch_uid="requires_consent_on_pre_save")

    fix_null_action_items(apps)

    for model in models:
        model_cls = apps.get_model(app_label, model)
        model_cls.action_name = [
            action.name
            for action in site_action_items.registry.values()
            if action.get_reference_model()
            and action.get_reference_model().split(".")[1].lower() == model.lower()
        ][0]
        if model_cls.action_name:
            for obj in model_cls.objects.all():
                sys.stdout.write(f"fixing {model_cls.action_name} action_item for {obj}.\n")
                if not obj.action_item:
                    try:
                        obj.action_item_id = action_item_cls.objects.get(
                            action_identifier=obj.action_identifier
                        ).pk
                    except MultipleObjectsReturned as e:
                        qs = action_item_cls.objects.filter(
                            action_identifier=obj.action_identifier
                        ).order_by("created")
                        raise MultipleObjectsReturned(f"{e} {qs}.")
                    else:
                        obj.save_base(update_fields=["action_item_id"])
                else:
                    try:
                        obj.save()
                    except MultipleObjectsReturned:
                        sys.stdout.write(f"Failed to resave {repr(obj)}")


def fix_null_action_items(apps):
    action_item_cls = apps.get_model("edc_action_item", "ActionItem")
    try:
        action_items = action_item_cls.objects.filter(
            parent_action_identifier__isnull=False, parent_action_item__isnull=True
        )
    except FieldError as e:
        print(e)
    else:
        for action_item in action_items:
            if not action_item.parent_action_item and action_item.parent_action_identifier:
                parent_action_item = action_item_cls.objects.get(
                    action_identifier=action_item.parent_action_identifier
                )
                updated = (
                    action_item_cls.objects.filter(
                        parent_action_identifier=action_item.parent_action_identifier,
                        parent_action_item__isnull=True,
                        linked_to_reference=True,
                    )
                    .exclude(action_identifier=action_item.related_action_identifier)
                    .update(parent_action_item=parent_action_item)
                )
                sys.stdout.write(
                    f" setting parent_action_items for {parent_action_item}. Got {updated}\n"
                )
    try:
        action_items = action_item_cls.objects.filter(
            related_action_identifier__isnull=False, related_action_item__isnull=True
        )
    except FieldError as e:
        print(e)
    else:
        for action_item in action_items:
            if not action_item.related_action_item and action_item.related_action_identifier:
                related_action_item = action_item_cls.objects.get(
                    action_identifier=action_item.related_action_identifier
                )
                updated = (
                    action_item_cls.objects.filter(
                        related_action_identifier=action_item.related_action_identifier,
                        related_action_item__isnull=True,
                    )
                    .exclude(action_identifier=action_item.related_action_identifier)
                    .update(related_action_item=related_action_item)
                )
                sys.stdout.write(
                    f" setting related_action_items for "
                    f"{related_action_item}. Got {updated}\n"
                )


def fix_null_related_action_items(apps):  # noqa
    """"""
    post_save.disconnect(dispatch_uid="serialize_on_save")
    pre_save.disconnect(dispatch_uid="requires_consent_on_pre_save")
    action_item_cls = apps.get_model("edc_action_item", "ActionItem")

    fix_null_action_items(apps)

    related_action_items = {}
    for action_cls in site_action_items.registry.values():
        if action_cls.related_reference_fk_attr:
            for action_item in action_item_cls.objects.filter(
                related_action_item__isnull=True, action_type__name=action_cls.name
            ):
                related_action_item = None
                reference_obj = None
                try:
                    reference_obj = action_item.reference_obj
                except ObjectDoesNotExist:
                    print("No reference object", action_item)
                else:
                    try:
                        related_reference_obj = getattr(
                            reference_obj, action_cls.related_reference_fk_attr
                        )
                    except ObjectDoesNotExist:
                        print("related_reference_obj does not exist")
                        if action_item.parent_action_item:
                            related_action_item = action_item.parent_action_item
                        elif reference_obj.parent_action_item:
                            related_action_item = reference_obj.parent_action_item
                    else:
                        related_action_item = related_reference_obj.action_item
                if related_action_item:
                    related_action_items.update({related_action_item: related_action_item})
                    action_item.related_action_item = related_action_item
                    action_item.save()
                    if reference_obj:
                        reference_obj.related_action_item = related_action_item  # noqa
                        reference_obj.save()
            if (
                action_item_cls.objects.filter(
                    related_action_item__isnull=True, action_type__name=action_cls.name
                ).count()
                > 0
            ):
                print("Some related action identifiers are still `none`")

    # verify sequence
    for related_action_item in related_action_items:
        fix_action_item_sequence(
            action_item_cls,
            action_identifier=related_action_item.action_identifier,
            subject_identifier=related_action_item.subject_identifier,
        )


def fix_action_item_sequence(action_item_cls, action_identifier=None, subject_identifier=None):
    """Verify the parent sequence of action items given the
    related action item.
    """
    parent_action_item = action_item_cls.objects.get(
        subject_identifier=subject_identifier, action_identifier=action_identifier
    )
    last_action_item = None
    for action_item in action_item_cls.objects.filter(
        related_action_item=parent_action_item
    ).order_by("created"):
        action_item.parent_action_item = last_action_item or parent_action_item
        action_item.save()
        action_item.refresh_from_db()
        if action_item.status in [CLOSED, OPEN]:
            action_item.reference_obj.parent_action_item = parent_action_item
            action_item.reference_obj.save()
            action_item.reference_obj.refresh_from_db()
        last_action_item = action_item


def fix_duplicate_singleton_action_items(apps, name=None):
    post_delete.disconnect(dispatch_uid="serialize_on_post_delete")
    post_save.disconnect(dispatch_uid="serialize_on_save")
    pre_save.disconnect(dispatch_uid="requires_consent_on_pre_save")

    registered_subject_cls = apps.get_model("edc_registration", "RegisteredSubject")
    action_item_cls = apps.get_model("edc_action_item", "ActionItem")

    action_cls = site_action_items.get(name)
    if action_cls.singleton:
        for registered_subject in registered_subject_cls.objects.all():
            try:
                action_item_cls.objects.get(
                    subject_identifier=registered_subject.subject_identifier,
                    action_type__name=name,
                )
            except ObjectDoesNotExist:
                pass
            except MultipleObjectsReturned:
                try:
                    action_item_cls.objects.get(
                        subject_identifier=registered_subject.subject_identifier,
                        action_type__name=name,
                        status=CLOSED,
                    )
                except (ObjectDoesNotExist, MultipleObjectsReturned):
                    for index, action_item in enumerate(
                        action_item_cls.objects.filter(
                            subject_identifier=registered_subject.subject_identifier,
                            action_type__name=name,
                        )
                    ):
                        if index > 0:
                            action_item.delete()
                else:
                    action_item_cls.objects.filter(
                        subject_identifier=registered_subject.subject_identifier,
                        action_type__name=name,
                        status=NEW,
                    ).delete()


def fix_null_related_action_items2(delete_orphans=None):  # noqa
    """Used for early action items where related_action_item
    was not set, e.g. v0.1.1

    Call from shell.

    mysql:
        select action_identifier, subject_identifier,
        created from edc_action_item_actionitem
        where related_action_item_id is NULL
        and reference_model = 'ambition_ae.aefollowup';
    """
    from django.apps import apps as django_apps

    post_save.disconnect(dispatch_uid="serialize_on_save")
    pre_save.disconnect(dispatch_uid="requires_consent_on_pre_save")
    action_item_cls = django_apps.get_model("edc_action_item", "ActionItem")
    for action_cls in site_action_items.registry.values():
        if action_cls.related_reference_fk_attr:
            for action_item in action_item_cls.objects.filter(
                related_action_item__isnull=True, action_type__name=action_cls.name
            ):
                try:
                    action_item.reference_obj
                except ObjectDoesNotExist as e:
                    if delete_orphans:
                        print(f"Deleting orphaned action item {action_item}.")
                        action_item.delete()
                    else:
                        print(f"Skipping {action_item}. Got {e}")
                else:
                    print(
                        action_item,
                        action_cls.related_reference_fk_attr,
                        action_item.reference_obj,
                    )
                    action_item.related_action_item = getattr(
                        action_item.reference_obj, action_cls.related_reference_fk_attr
                    ).action_item
                    action_item.save()



from typing import List

from django.db import models

from .action import Action
from .action_item_notification import ActionItemNotification


class ActionWithNotification(Action):

    """A class mixin for the Action class that adds
    support for notifications.
    """

    notification_email_to: List[str] = None
    notification_display_name: str = None
    notification_fields: List[str] = None
    notification_super_cls = ActionItemNotification
    notify_on_changed_reference_obj: models.Model = True
    notify_on_close: bool = False
    notify_on_new: bool = False
    notify_on_new_and_no_reference_obj: bool = True
    notify_on_open: bool = False

    @classmethod
    def notification_cls(cls):
        """Returns a subclass of ActionItemModelNotification."""
        return type(
            f"{cls.__name__}Notification",
            (cls.notification_super_cls,),
            dict(
                name=f"{cls.name}-notification",
                notification_action_name=cls.name,
                display_name=(
                    cls.notification_display_name or f"{cls.display_name} Notification"
                ),
                email_to=cls.notification_email_to,
                notification_fields=cls.notification_fields,
                model=cls.get_reference_model(),
                notify_on_new_and_no_reference_obj=cls.notify_on_new_and_no_reference_obj,
                notify_on_new=cls.notify_on_new,
                notify_on_open=cls.notify_on_open,
                notify_on_close=cls.notify_on_close,
                notify_on_changed_reference_obj=cls.notify_on_changed_reference_obj,
            ),
        )



from unittest.case import skip

from django.core import mail
from django.test import TestCase
from edc_constants.constants import NEW
from edc_notification import NewModelNotification, UpdatedModelNotification
from edc_notification.site_notifications import site_notifications

from edc_action_item.action_item_notification import (
    NOTIFY_ON_CHANGED_REFERENCE_OBJ,
    NOTIFY_ON_NEW_AND_NO_REFERENCE_OBJ,
)
from edc_action_item.site_action_items import site_action_items

from ..action_items import FormZeroAction, register_actions
from ..models import FormZero
from ..test_case_mixin import TestCaseMixin


class TestActionNotification(TestCaseMixin, TestCase):
    def setUp(self):
        register_actions()
        self.subject_identifier = self.fake_enroll()

    def test_sends_correct_number_of_emails(self):
        self.assertIn(FormZeroAction, site_action_items.registry.values())

        # action without reference obj
        form_zero_action = FormZeroAction(subject_identifier=self.subject_identifier)
        self.assertEqual(len(mail.outbox), 1)
        self.assertIn(NOTIFY_ON_NEW_AND_NO_REFERENCE_OBJ, mail.outbox[0].body)
        self.assertEqual(form_zero_action.action_item.status, NEW)

        # action with reference obj
        form_zero = FormZero.objects.create(subject_identifier=self.subject_identifier, f1="1")
        form_zero.refresh_from_db()
        self.assertEqual(len(mail.outbox), 1)

        # change/update field on reference obj that is a notification field
        form_zero.f1 = "2"
        form_zero.save()
        form_zero.refresh_from_db()
        self.assertEqual(len(mail.outbox), 2)
        self.assertIn("*UPDATE*", mail.outbox[1].subject)
        self.assertIn("An updated report", mail.outbox[1].body)
        self.assertIn(NOTIFY_ON_CHANGED_REFERENCE_OBJ, mail.outbox[1].body)

        # resave reference obj without any change
        form_zero.save()
        form_zero.refresh_from_db()
        self.assertEqual(len(mail.outbox), 2)

        form_zero.f1 = "1"
        form_zero.save()
        form_zero.refresh_from_db()
        self.assertEqual(len(mail.outbox), 3)
        self.assertIn(NOTIFY_ON_CHANGED_REFERENCE_OBJ, mail.outbox[1].body)

    @skip("notification")
    def test_notification_registration(self):
        """Asserts that by registering the action class, the
        notification class is also registered.
        """
        action_enabled_for_notifications = []
        for action_cls in site_action_items.registry.values():
            action_enabled_for_notifications.append(action_cls)

        self.assertIsNotNone(action_enabled_for_notifications)

        # assert both the new and update are registered
        for action_cls in action_enabled_for_notifications:
            self.assertIn(
                action_cls.notification_name or action_cls.name,
                site_notifications.registry,
            )
            self.assertIn(
                f"{action_cls.notification_name or action_cls.name}_update",
                site_notifications.registry,
            )

        # assert class type for new and update
        for k, v in site_notifications.registry.items():
            if "update" in k:
                self.assertTrue(issubclass(v, UpdatedModelNotification), msg=v.name)
            else:
                self.assertTrue(issubclass(v, NewModelNotification), msg=v.name)

    def test_action_sends_notification_on_new(self):
        FormZero.objects.create(subject_identifier=self.subject_identifier)
        self.assertEqual(len(mail.outbox), 1)
        self.assertNotIn("*UPDATE*", mail.outbox[0].subject)

    def test_action_sends_notification_does_not_duplicate_send(self):
        form_zero = FormZero.objects.create(subject_identifier=self.subject_identifier)
        self.assertEqual(len(mail.outbox), 1)
        form_zero.save()
        self.assertEqual(len(mail.outbox), 1)

    def test_action_sends_another_notification_on_update(self):
        form_zero = FormZero.objects.create(f1=1, subject_identifier=self.subject_identifier)
        self.assertEqual(len(mail.outbox), 1)
        form_zero.f1 = 2
        form_zero.save()
        self.assertEqual(len(mail.outbox), 2)
        form_zero.save()
        self.assertEqual(len(mail.outbox), 2)
        self.assertIn("*UPDATE*", mail.outbox[1].subject)

    def test_notification_updates_the_actions_model_as_emailed(self):
        form_zero = FormZero.objects.create(subject_identifier=self.subject_identifier)
        form_zero.refresh_from_db()
        self.assertTrue(form_zero.action_item.emailed)
        self.assertIsNotNone(form_zero.action_item.emailed_datetime)

    def test_action_sends_as_test_email(self):
        FormZero.objects.create(subject_identifier=self.subject_identifier)
        self.assertEqual(len(mail.outbox), 1)
        self.assertIn("TEST/UAT", mail.outbox[0].subject)
        self.assertIn("THIS IS A TEST MESSAGE", mail.outbox[0].body)

    def test_action_sends_as_test_email_with_update(self):
        form_zero = FormZero.objects.create(f1=1, subject_identifier=self.subject_identifier)
        self.assertEqual(len(mail.outbox), 1)
        self.assertNotIn("*UPDATE*", mail.outbox[0].subject)
        self.assertIn("A report", mail.outbox[0].body)
        form_zero.f1 = 2
        form_zero.save()
        self.assertEqual(len(mail.outbox), 2)
        self.assertIn("*UPDATE*", mail.outbox[1].subject)
        self.assertIn("An updated report", mail.outbox[1].body)



from dateutil.relativedelta import relativedelta
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit import Crf, CrfCollection, Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

crfs = CrfCollection(
    Crf(show_order=1, model="edc_action_item.crflongitudinalone", required=True),
    Crf(show_order=2, model="edc_action_item.crflongitudinaltwo", required=True),
)

visit0 = Visit(
    code="1000",
    title="Day 1",
    timepoint=0,
    rbase=relativedelta(days=0),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    requisitions=None,
    crfs=crfs,
    crfs_unscheduled=None,
    requisitions_unscheduled=None,
    facility_name="7-day-clinic",
)

visit1 = Visit(
    code="2000",
    title="Day 2",
    timepoint=1,
    rbase=relativedelta(days=1),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    requisitions=None,
    crfs=crfs,
    crfs_unscheduled=None,
    requisitions_unscheduled=None,
    facility_name="7-day-clinic",
)


schedule = Schedule(
    name="schedule",
    onschedule_model="edc_metadata.onschedule",
    offschedule_model="edc_metadata.offschedule",
    consent_model="edc_metadata.subjectconsent",
    appointment_model="edc_appointment.appointment",
)

schedule.add_visit(visit0)
schedule.add_visit(visit1)

visit_schedule = VisitSchedule(
    name="visit_schedule",
    offstudy_model="edc_offstudy.subjectoffstudy",
    death_report_model="edc_metadata.deathreport",
)

visit_schedule.add_schedule(schedule)



from django.core.exceptions import ObjectDoesNotExist

from .create_or_update_action_type import create_or_update_action_type


def get_action_type(cls, name=None, using=None):
    """Returns the ActionType model instance."""
    try:
        action_type = (
            cls.action_type_model_cls().objects.using(using).get(name=name or cls.name)
        )
    except ObjectDoesNotExist:
        action_type = create_or_update_action_type(using=using, **cls.as_dict())
    return action_type



from django.core.checks import Warning, register

from edc_action_item.site_action_items import site_action_items


@register()
def edc_notification_check(app_configs, **kwargs):  # noqa
    errors = []

    for action_cls in site_action_items.registry.items():
        try:
            action_cls.reference_model_cls().history
        except AttributeError:
            errors.append(
                Warning(
                    (
                        f"Reference model used by action mcs {action_cls} "
                        f"has no history manager."
                    ),
                    hint="History manager is need to detect changes.",
                    obj=action_cls,
                    id="edc_action_item.E001",
                )
            )
    return errors



from django.apps import apps as django_apps
from django.db.models.signals import post_save
from django.test import TestCase

from edc_action_item.data_fixers import (
    fix_null_action_item_fk,
    fix_null_related_action_items,
)
from edc_action_item.models import update_or_create_action_item_on_post_save

from ..action_items import register_actions
from ..models import FormOne, FormTwo
from ..test_case_mixin import TestCaseMixin


class TestUtils(TestCaseMixin, TestCase):
    def setUp(self):
        register_actions()
        self.subject_identifier = self.fake_enroll()

    def test_fix_null_related_action_items(self):
        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        form_two = FormTwo.objects.create(
            subject_identifier=self.subject_identifier, form_one=form_one
        )
        form_two.refresh_from_db()

        self.assertEqual(form_one.action_item, form_two.related_action_item)

        # break it ....
        # set related_action_item to None
        form_two.related_action_item = None
        form_two.save_base()
        form_two.refresh_from_db()
        form_two.action_item.related_action_item = None
        form_two.action_item.save_base()
        form_two.action_item.refresh_from_db()

        # it's broken ...
        # assert related_action_items are None
        self.assertIsNone(form_two.related_action_item)
        self.assertIsNone(form_two.action_item.related_action_item)

        # fix
        fix_null_related_action_items(django_apps)

        # assert related_action_item are no longer None
        form_two.refresh_from_db()
        form_two.action_item.refresh_from_db()

        self.assertEqual(form_one.action_item, form_two.action_item.related_action_item)

    def test_fix_null_related_action_items2(self):
        fix_null_related_action_items(django_apps)

        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        form_two_a = FormTwo.objects.create(
            subject_identifier=self.subject_identifier, form_one=form_one
        )
        form_two_b = FormTwo.objects.create(
            subject_identifier=self.subject_identifier, form_one=form_one
        )

        self.assertEqual(form_one.action_item, form_two_b.related_action_item)

        # set related_action_item to None
        form_two_b.related_action_item = None
        form_two_b.action_item.related_action_item = None
        post_save.disconnect(dispatch_uid="update_or_create_action_item_on_post_save")
        form_two_b.action_item.save()
        form_two_b.save_base()
        post_save.connect(
            update_or_create_action_item_on_post_save,
            dispatch_uid="update_or_create_action_item_on_post_save",
        )

        form_two_b.refresh_from_db()
        form_two_b.action_item.refresh_from_db()

        # assert related_action_items are None
        self.assertIsNone(form_two_b.related_action_item)
        self.assertIsNone(form_two_b.action_item.related_action_item)

        # fix
        fix_null_related_action_items(django_apps)

        # assert related_action_item are NOT None
        form_two_b.refresh_from_db()
        self.assertEqual(form_one.action_item, form_two_b.action_item.related_action_item)
        self.assertEqual(form_one.action_item, form_two_b.related_action_item)
        self.assertEqual(form_two_a.action_item, form_two_b.parent_action_item)

    def test_fix_null_related_action_items3(self):
        # fix_null_related_action_items(django_apps)

        # form_one -> form_two_a -> form_two_b

        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        form_two_a = FormTwo.objects.create(
            subject_identifier=self.subject_identifier, form_one=form_one
        )
        form_two_b = FormTwo.objects.create(
            subject_identifier=self.subject_identifier, form_one=form_one
        )

        # set related_action_item to None
        form_two_a.parent_action_item = form_one.action_item
        form_two_a.action_item.parent_action_item = form_one.action_item

        form_two_a.action_item.related_action_item = None
        form_two_a.related_action_item = None
        form_two_a.action_item.save_base()

        post_save.disconnect(dispatch_uid="update_or_create_action_item_on_post_save")
        form_two_a.save_base()
        form_two_a.refresh_from_db()

        form_two_b.related_action_item = None
        form_two_b.action_item.related_action_item = None
        form_two_b.action_item.save_base()
        form_two_b.save_base()
        form_two_b.refresh_from_db()

        # set parent b to wrong parent
        form_two_b.parent_action_item = form_one.action_item
        form_two_b.action_item.parent_action_item = form_one.action_item
        form_two_b.action_item.save_base()
        form_two_b.save_base()
        form_two_b.refresh_from_db()
        post_save.connect(
            update_or_create_action_item_on_post_save,
            dispatch_uid="update_or_create_action_item_on_post_save",
        )

        self.assertEqual(form_one.action_item, form_two_b.parent_action_item)
        self.assertEqual(form_one.action_item, form_two_b.action_item.parent_action_item)

        # fix
        fix_null_related_action_items(django_apps)

        form_two_a.refresh_from_db()
        form_two_a.action_item.refresh_from_db()
        form_two_a.parent_action_item.refresh_from_db()
        form_two_b.refresh_from_db()
        form_two_b.action_item.refresh_from_db()
        form_two_b.parent_action_item.refresh_from_db()

        # assert related_action_item are NOT None
        self.assertEqual(form_one.action_item, form_two_a.action_item.related_action_item)
        self.assertEqual(form_one.action_item, form_two_b.related_action_item)

        self.assertEqual(form_two_a.parent_action_item, form_two_a.related_action_item)

        self.assertEqual(form_one.action_item, form_two_a.parent_action_item)

        self.assertEqual(form_one.action_item, form_two_a.action_item.parent_action_item)

        # assert parent was fixed
        self.assertEqual(form_two_a.action_item, form_two_b.action_item.parent_action_item)
        self.assertEqual(form_two_a.action_item, form_two_b.parent_action_item)

        self.assertNotEqual(form_two_b.parent_action_item, form_two_b.related_action_item)

    def test_fix_null_action_item_fk(self):
        form_one = FormOne.objects.create(subject_identifier=self.subject_identifier)
        FormTwo.objects.create(subject_identifier=self.subject_identifier, form_one=form_one)
        FormTwo.objects.create(subject_identifier=self.subject_identifier, form_one=form_one)
        fix_null_action_item_fk(
            django_apps, app_label="edc_action_item", models=["formone", "formtwo"]
        )



import sys

from django.conf import settings

from .action_item import ActionItem
from .action_model_mixin import ActionModelMixin, ActionNoManagersModelMixin
from .action_type import ActionType
from .edc_permissions import EdcPermissions
from .reference import Reference
from .signals import (
    action_item_notification_on_post_create_historical_record,
    action_on_reference_model_post_delete,
    update_action_item_reason_on_m2m_changed,
    update_or_create_action_item_on_m2m_change,
    update_or_create_action_item_on_post_save,
)

if (
    settings.APP_NAME == "edc_action_item"
    and "migrate" not in sys.argv
    and "makemigrations" not in sys.argv
):
    from ..tests import models



# Generated by Django 2.0.7 on 2018-07-07 14:59

import _socket
import django.db.models.deletion
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import edc_utils
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_action_item", "0005_auto_20180409_1005")]

    operations = [
        migrations.CreateModel(
            name="Action",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("tracking_identifier", models.CharField(max_length=30, null=True)),
                ("subject_identifier", models.CharField(max_length=50)),
                (
                    "parent_reference_identifier",
                    models.CharField(max_length=30, null=True),
                ),
                (
                    "related_reference_identifier",
                    models.CharField(max_length=30, null=True),
                ),
                ("action_identifier", models.CharField(max_length=25, unique=True)),
                (
                    "report_datetime",
                    models.DateTimeField(default=edc_utils.date.get_utcnow),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.AlterModelOptions(
            name="historicalactionitem",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Action Item",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalactionitemupdate",
            options={
                "get_latest_by": "history_date",
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical action item update",
            },
        ),
        migrations.RenameField(
            model_name="actiontype", old_name="model", new_name="reference_model"
        ),
        migrations.AddField(
            model_name="actionitem",
            name="linked_to_reference",
            field=models.BooleanField(default=False, editable=False),
        ),
        migrations.AddField(
            model_name="historicalactionitem",
            name="linked_to_reference",
            field=models.BooleanField(default=False, editable=False),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="parent_reference_identifier",
            field=models.CharField(
                blank=True,
                help_text="May be left blank. e.g. action identifier from reference model that opened the item (parent).",
                max_length=50,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="reference_identifier",
            field=models.CharField(
                help_text="e.g. action identifier updated from the reference model",
                max_length=50,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="related_reference_identifier",
            field=models.CharField(
                blank=True,
                help_text="May be left blank. e.g. action identifier from source model that opened the item.",
                max_length=50,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="parent_action_item",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_action_item.ActionItem",
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="parent_reference_identifier",
            field=models.CharField(
                blank=True,
                help_text="May be left blank. e.g. action identifier from reference model that opened the item (parent).",
                max_length=50,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="reference_identifier",
            field=models.CharField(
                help_text="e.g. action identifier updated from the reference model",
                max_length=50,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="related_reference_identifier",
            field=models.CharField(
                blank=True,
                help_text="May be left blank. e.g. action identifier from source model that opened the item.",
                max_length=50,
                null=True,
            ),
        ),
        migrations.AlterUniqueTogether(
            name="actionitem",
            unique_together={("subject_identifier", "action_type", "action_identifier")},
        ),
        migrations.AddField(
            model_name="action",
            name="action_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="action",
                to="edc_action_item.ActionType",
                verbose_name="Action",
            ),
        ),
    ]



from edc_constants.constants import NEW, OPEN

from ..model_wrappers import ActionItemModelWrapper
from ..models import ActionItem


class ActionItemViewMixin:
    action_item_model_wrapper_cls = ActionItemModelWrapper

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(open_action_items=self.open_action_items)
        return context

    @property
    def open_action_items(self):
        """Returns a list of wrapped ActionItem instances
        where status is NEW or OPEN.
        """
        qs = ActionItem.on_site.filter(
            subject_identifier=self.kwargs.get("subject_identifier"),
            action_type__show_on_dashboard=True,
            status__in=[NEW, OPEN],
        ).order_by("-report_datetime")
        return [self.action_item_model_wrapper_cls(model_obj=obj) for obj in qs]
